import { Broker, wsQueryRequest } from '@/lib/data-access/broker';
import {
  queryingBackend,
  useActiveQuery,
  useActiveSaveState,
  useAppDispatch,
  useConfig,
  useML,
  useQuerybuilderHash,
  useSchemaGraph,
  useSearchResultQB,
} from '@/lib/data-access/store';
import { setQuerybuilderGraphology, setQuerybuilderSettings, toQuerybuilderGraphology } from '@/lib/data-access/store/sessionSlice';
import html2canvas from 'html2canvas';
import React, { useCallback, useEffect, useMemo, useRef, useState } from 'react';
import { Node, ReactFlowInstance, ReactFlowProvider, useReactFlow } from 'reactflow';
import { QueryElementTypes, QueryViewMode, defaultGraph } from 'ts-common';
import { Dialog, DialogContent } from '../../components/layout/Dialog';
import { addError } from '../../data-access/store/configSlice';
import { toSchemaGraphology } from '../../data-access/store/schemaSlice';
import { LayoutFactory } from '../../graph-layout';
import { SchemaReactflowLogicNode, createReactFlowElements, toHandleData } from '../model';

import { QueryBuilderContextMenu } from './QueryBuilderContextMenu';
import { QueryBuilderDispatcherContext } from './QueryBuilderDispatcher';
import { QueryInputCypher } from './querybuilderinputpanels/QueryInputCypher';
import { QueryInputVisual } from './querybuilderinputpanels/QueryInputVisual';
import { QueryBuilderNav, QueryBuilderToggleSettings } from './querybuildernav/QueryBuilderNav';
import { QueryBuilderLogicPillsPanel } from './querysidepanel/QueryBuilderLogicPillsPanel';
import { QueryBuilderRelatedNodesPanel } from './querysidepanel/QueryBuilderRelatedNodesPanel';
import { ConnectingNodeDataI } from './utils/connectorDrop';

export type QueryBuilderProps = {
  onRunQuery?: (useCached: boolean) => void;
};

/**
 * This is the main querybuilder component. It is responsible for holding all pills and fire off the visual part of the querybuilder panel logic
 */
export const QueryBuilderInner = (props: QueryBuilderProps) => {
  const [toggleSettings, setToggleSettings] = useState<QueryBuilderToggleSettings>();
  const reactFlowWrapper = useRef<HTMLDivElement | null>(null);
  const reactFlowRef = useRef<HTMLDivElement | null>(null);
  const ss = useActiveSaveState();
  const activeQuery = useActiveQuery();

  const schemaGraph = useSchemaGraph();
  const schema = useMemo(() => toSchemaGraphology(schemaGraph), [schemaGraph]);
  const qbHash = useQuerybuilderHash();
  const config = useConfig();
  const dispatch = useAppDispatch();
  const ml = useML();
  const connectingNodeId = useRef<ConnectingNodeDataI | null>(null);
  const editLogicNode = useRef<SchemaReactflowLogicNode | undefined>(undefined);
  const reactFlow = useReactFlow();

  const graphologyGraph = useMemo(() => toQuerybuilderGraphology(activeQuery?.graph ?? defaultGraph()), [activeQuery?.graph, qbHash]);
  const elements = useMemo(() => createReactFlowElements(graphologyGraph), [activeQuery?.graph, qbHash]);
  const searchResults = useSearchResultQB();
  const reactFlowInstanceRef = useRef<ReactFlowInstance | null>(null);

  const [contextMenuOpen, setContextMenuOpen] = useState<{
    open: boolean;
    node?: Node;
    position?: { x: number; y: number };
  }>({
    open: false,
  });

  /**
   * Highlights nodes that are present in the search results by setting their 'selected' attribute
   */
  const highlightSearchResults = useCallback(() => {
    const searchResultKeys = new Set([...searchResults.nodes.map(node => node.key), ...searchResults.edges.map(edge => edge.key)]);

    elements.nodes.forEach(node => {
      const isNodeInSearchResults = searchResultKeys.has(node?.id);
      if (isNodeInSearchResults && !graphologyGraph.getNodeAttribute(node?.id, 'selected')) {
        graphologyGraph.setNodeAttribute(node?.id, 'selected', true);
      } else if (!isNodeInSearchResults && graphologyGraph.getNodeAttribute(node?.id, 'selected')) {
        graphologyGraph.setNodeAttribute(node?.id, 'selected', false);
      }
    });

    dispatch(setQuerybuilderGraphology(graphologyGraph));
  }, [searchResults, elements.nodes, graphologyGraph]);

  useEffect(() => {
    highlightSearchResults();
  }, [searchResults]);

  /**
   * Clears all nodes in the graph.
   *  TODO: only works if the node is clicked and not moved (maybe use onSelectionChange)
   */
  function onNodesDelete(nodes: Node[]) {
    if (!ss?.authorization.query.W) return;
    nodes.forEach(n => {
      graphologyGraph.dropNode(n.id);
    });

    // Dispatch the new graphology object, so reactflow will get re-rendered
    dispatch(setQuerybuilderGraphology(graphologyGraph));
  }

  function applyLayout() {
    if (!activeQuery) return;
    if (activeQuery.settings.layout !== 'manual') {
      const factory = new LayoutFactory();
      factory.createLayout(activeQuery.settings.layout).layout(graphologyGraph);
      dispatch(setQuerybuilderGraphology(graphologyGraph));
    }
  }

  const fitView = () => {
    if (reactFlowInstanceRef.current) {
      reactFlowInstanceRef.current.fitView();
    }
  };

  const onNodeContextMenu = (event: React.MouseEvent, node: Node) => {
    if (event.shiftKey) return;
    event.preventDefault();

    // Get the bounding rectangle of the ReactFlow wrapper
    const reactFlowBounds = reactFlowWrapper.current?.getBoundingClientRect();

    // Calculate position relative to the ReactFlow container
    const position = reactFlowBounds
      ? {
          x: event.clientX + reactFlowBounds.left,
          y: event.clientY + reactFlowBounds.top,
        }
      : { x: event.clientX, y: event.clientY };

    setContextMenuOpen({ open: true, node: node, position });
  };

  const onMouseDown = useCallback(
    (event: React.MouseEvent | React.TouchEvent) => {
      if (contextMenuOpen.open) {
        setContextMenuOpen({ ...contextMenuOpen, open: false });
      }
    },
    [contextMenuOpen],
  );

  useEffect(() => {
    try {
      applyLayout();
    } catch (e) {
      console.error(e);
    }
    setContextMenuOpen({ ...contextMenuOpen, open: false });
  }, [activeQuery?.settings]);

  const handleScreenshot = async () => {
    if (reactFlowRef.current) {
      try {
        const canvas = await html2canvas(reactFlowRef.current);
        const screenshotUrl = canvas.toDataURL('image/jpeg');
        const link = document.createElement('a');
        link.href = screenshotUrl;
        link.download = 'querybuilder.png';
        link.click();
      } catch (error) {
        dispatch(addError('Screenshot failed.'));
      }
    }
  };

  const handleManualRunQuery = () => {
    if (ss == null || activeQuery == null || activeQuery.id == null || activeQuery.settings.cypher == null) return;

    const callID = Broker.instance().generateCallID();
    dispatch(queryingBackend({ callID, status: 'querying' }));
    wsQueryRequest({
      saveStateID: ss.id,
      queryID: activeQuery.id,
      ml,
      manualQuery: activeQuery.settings.cypher,
      callID,
      useCached: false,
    });
  };

  const onAddSchemaToQueryBuilder = () => {
    if (!activeQuery) return;

    // TODO: incomplete (not being shown) for now due to lack of "OPTIONAL MATCH" support in the backend

    const addedNodes: Record<string, string> = {};
    // loop through all nodes in the schema and add them to the graph
    schemaGraph.nodes.forEach(node => {
      if (!node.key) return;

      const { id } = graphologyGraph.addPill2Graphology(
        {
          type: QueryElementTypes.Entity,
          // x: position.x - mouse_x,
          // y: position.y - mouse_y,
          x: 0,
          y: 0,
          name: node.key,
          schemaKey: node.key,
          attributes: [],
        },
        schema.getNodeAttribute(node.key, 'attributes'),
      );
      if (!id) return;
      addedNodes[node.key] = id;
    });

    // loop all edges in the schema and add them to the graph
    schemaGraph.edges.forEach(edge => {
      if (!edge.key || !edge.source || !edge.target) return;
      const fromNode = graphologyGraph.getNodeAttributes(addedNodes[edge.source]);
      const toNode = graphologyGraph.getNodeAttributes(addedNodes[edge.target]);
      if (!fromNode || !toNode) return;

      const relation = graphologyGraph.addPill2Graphology(
        {
          type: QueryElementTypes.Relation,
          // x: position.x,
          // y: position.y,
          x: 0,
          y: 0,
          depth: { min: activeQuery.settings.depth.min, max: activeQuery.settings.depth.max },
          name: edge.attributes?.collection,
          schemaKey: edge.attributes?.name,
          collection: edge.attributes?.collection,
          attributes: [],
        },
        schema.getEdgeAttribute(edge.key, 'attributes'),
      );

      graphologyGraph.addEdge2Graphology(fromNode, relation, {
        type: 'connection',
        sourceHandleData: toHandleData(edge.source),
        targetHandleData: toHandleData(edge.key),
      });

      graphologyGraph.addEdge2Graphology(relation, toNode, {
        type: 'connection',
        sourceHandleData: toHandleData(edge.key),
        targetHandleData: toHandleData(edge.target),
      });
    });

    dispatch(setQuerybuilderGraphology(graphologyGraph));
  };

  return (
    <QueryBuilderDispatcherContext.Provider
      value={{
        openLogicPillUpdate: node => {
          editLogicNode.current = node;
          setToggleSettings('logic');
        },
        openLogicPillCreate: (params, position) => {
          if (!params?.handleId) return;

          const node = graphologyGraph.getNodeAttributes(params.nodeId);
          const handleData = toHandleData(params.handleId);
          connectingNodeId.current = {
            params,
            node,
            position: position || { x: 0, y: 0 },
            attribute: { handleData: handleData },
          };
          setToggleSettings('logic');
        },
      }}
    >
      <QueryBuilderContextMenu
        open={contextMenuOpen.open}
        node={contextMenuOpen.node}
        position={contextMenuOpen.position}
        reactFlowWrapper={reactFlowWrapper}
        reactFlow={reactFlow}
        onClose={() => {
          setContextMenuOpen({ ...contextMenuOpen, open: false });
        }}
      />
      <div ref={reactFlowWrapper} className="h-full w-full flex flex-col relative">
        <QueryBuilderNav
          toggleSettings={toggleSettings}
          onFitView={fitView}
          onRunQuery={(useCached: boolean) => {
            if (props.onRunQuery) props.onRunQuery(useCached);
          }}
          onScreenshot={handleScreenshot}
          onLogic={() => {
            if (toggleSettings === 'logic') setToggleSettings(undefined);
            else setToggleSettings('logic');
          }}
        />

        <Dialog
          open={toggleSettings === 'logic'}
          onOpenChange={ret => {
            if (!ret) setToggleSettings(undefined);
          }}
        >
          <DialogContent className="w-fit rounded-none">
            <QueryBuilderLogicPillsPanel
              onClick={v => {
                connectingNodeId.current = null;
                editLogicNode.current = undefined;
                setToggleSettings(undefined);
              }}
              reactFlowWrapper={reactFlowWrapper.current}
              title="Available Logic Pills"
              className="max-h-[75vh] min-h-[45vh]"
              onDrag={() => {}}
              connection={connectingNodeId?.current}
              editNode={editLogicNode.current}
            />
          </DialogContent>
        </Dialog>
        <Dialog
          open={toggleSettings === 'relatedNodes'}
          onOpenChange={ret => {
            if (!ret) setToggleSettings(undefined);
          }}
        >
          <DialogContent>
            <QueryBuilderRelatedNodesPanel
              onFinished={() => {
                connectingNodeId.current = null;
                setToggleSettings(undefined);
              }}
              reactFlowWrapper={reactFlowWrapper.current}
              title="Related nodes available to add to the query"
              className="min-h-[75vh] max-h-[75vh]"
              connection={connectingNodeId?.current}
            />
          </DialogContent>
        </Dialog>
        {activeQuery?.settings.mode == QueryViewMode.Cypher ? (
          <QueryInputCypher
            cypher={activeQuery?.settings.cypher}
            onCypherChange={newCypherString => {
              // console.log('Cypher changed', newCypherString);
              dispatch(setQuerybuilderSettings({ settings: { cypher: newCypherString } }));
            }}
            onRunQuery={() => {
              handleManualRunQuery();
            }}
          />
        ) : (
          <QueryInputVisual
            reactFlowRef={reactFlowRef}
            reactFlowWrapper={reactFlowWrapper}
            elements={elements}
            graphologyGraph={graphologyGraph}
            schema={schema}
            activeQuery={activeQuery}
            config={config}
            ss={ss}
            connectingNodeId={connectingNodeId}
            reactFlowInstanceRef={reactFlowInstanceRef}
            onSetToggleSettings={(settings?: string | undefined) => setToggleSettings(settings as QueryBuilderToggleSettings | undefined)}
            onMouseDown={onMouseDown}
            onNodeContextMenu={onNodeContextMenu}
            onAddSchemaToQueryBuilder={onAddSchemaToQueryBuilder}
          />
        )}
      </div>
    </QueryBuilderDispatcherContext.Provider>
  );
};

export const QueryBuilder = (props: QueryBuilderProps) => {
  return (
    <div className="query-panel flex w-full h-full border bg-light">
      <ReactFlowProvider>
        <QueryBuilderInner {...props} />
      </ReactFlowProvider>
    </div>
  );
};

export default QueryBuilder;
