import {
  Menu,
  MenuCheckboxItem,
  MenuContent,
  MenuItem,
  MenuLabel,
  MenuRadioGroup,
  MenuRadioItem,
  MenuSeparator,
  MenuTrigger,
} from '@/lib/components/menu';
import {
  attributeShownToggle,
  setQuerybuilderGraphology,
  setQueryUnionType,
  toQuerybuilderGraphology,
} from '@/lib/data-access/store/sessionSlice';
import { isEqual } from 'lodash-es';
import { useEffect, useMemo, useState } from 'react';
import { Node, ReactFlowInstance } from 'reactflow';
import { defaultGraph, NodeAttribute, QueryGraphEdgeHandle, QueryRelationDirection, QueryUnionType } from 'ts-common';
import { useActiveQuery, useAppDispatch, useQuerybuilderHash } from '../..';
import { getDataTypeIcon } from '../../components/DataTypeIcon';

export const QueryBuilderContextMenu = (props: {
  open: boolean;
  position?: { x: number; y: number };
  node?: Node;
  reactFlowWrapper: React.RefObject<HTMLDivElement | null>;
  reactFlow: ReactFlowInstance;
  onClose: () => void;
}) => {
  const dispatch = useAppDispatch();
  const activeQuery = useActiveQuery();
  const qbHash = useQuerybuilderHash();
  const unionType =
    activeQuery?.settings.unionTypes == null || props.node == null ? QueryUnionType.AND : activeQuery?.settings.unionTypes[props.node.id];
  const relationDirection =
    props.node && props.node.type === 'relation' ? (props.node.data?.direction ?? QueryRelationDirection.BOTH) : undefined;

  const graphologyGraph = useMemo(() => toQuerybuilderGraphology(activeQuery?.graph || defaultGraph()), [activeQuery?.graph, qbHash]);

  const [divPosition, setDivPosition] = useState<DOMRect>(new DOMRect());
  const [initialPosition, setInitialPosition] = useState<DOMRect>(new DOMRect());

  const state = useMemo(() => {
    if (!divPosition || !initialPosition || !props.node) return;
    let position = { x: 0, y: 0 };
    if (props.position) {
      position = props.position;
    } else {
      position = props.reactFlow.flowToScreenPosition({ x: props.node.data.x, y: props.node.data.y });
    }

    return {
      open: props.open,
      x: position.x - divPosition.x - (initialPosition.x - divPosition.x) + 10,
      y: position.y - divPosition.y - (initialPosition.y - divPosition.y) + 10,
    };
  }, [props.open, props.node, divPosition]);

  useEffect(() => {
    if (props.reactFlowWrapper.current == null) return;
    setInitialPosition(props.reactFlowWrapper.current.getBoundingClientRect());

    const resizeObserver = new ResizeObserver(() => {
      if (props.reactFlowWrapper.current == null) return;
      setDivPosition(props.reactFlowWrapper.current.getBoundingClientRect());
    });
    resizeObserver.observe(props.reactFlowWrapper.current);
    return () => {
      if (props.reactFlowWrapper.current == null) return;
      resizeObserver.disconnect(); // clean up
    };
  }, [props.reactFlowWrapper]);

  function isAttributeAdded(attribute: NodeAttribute): boolean {
    return activeQuery?.attributesBeingShown.some(x => isEqual(x, attribute.handleData)) ?? false;
  }

  function toggleAttribute(attribute: NodeAttribute) {
    dispatch(attributeShownToggle(attribute.handleData));
  }

  function setUnionType(value: string) {
    const unionType = value as QueryUnionType;
    if (!props.node) return;
    dispatch(setQueryUnionType({ nodeId: props.node.id, unionType: unionType }));
  }

  function setRelationDirection(value: string) {
    const relationDirection = value as QueryRelationDirection;
    if (!props.node) return;
    // @ts-expect-error - graphology types are not up to date
    graphologyGraph.setNodeAttribute(props.node.id, 'direction', relationDirection);
    dispatch(setQuerybuilderGraphology(graphologyGraph));
    props.onClose();
  }

  function removeNode() {
    if (!props.node) return;
    const connectedLogicPills = graphologyGraph.neighbors(props.node.id);
    connectedLogicPills.forEach(pill => {
      const attributes = graphologyGraph.getNodeAttributes(pill);
      if (attributes.type === 'logic') {
        graphologyGraph.dropNode(pill);
      }
    });

    graphologyGraph.dropNode(props.node.id);
    dispatch(setQuerybuilderGraphology(graphologyGraph));
    props.onClose();
  }

  return (
    <Menu isOpen={props.open && state !== undefined}>
      <MenuTrigger x={state ? state.x : 0} y={state ? state.y : 0}>
        <div className="hidden"></div>
      </MenuTrigger>
      <MenuContent>
        <Menu>
          <MenuTrigger label="Add/remove attribute" />
          <MenuContent className="max-h-60 overflow-y-auto">
            {props.node?.data.attributes.map((attr: { handleData: QueryGraphEdgeHandle }) => (
              <MenuCheckboxItem
                key={attr.handleData.attributeName + attr.handleData.nodeId}
                checked={isAttributeAdded(attr)}
                onCheckedChange={_ => toggleAttribute(attr)}
                onClick={e => {
                  if ((e.nativeEvent as PointerEvent)?.pointerId == -1) return; // Dont close menu when navigating with arrows and pressing enter
                  props.onClose();
                }}
                label={attr.handleData.attributeName}
                iconRight={getDataTypeIcon(attr?.handleData?.attributeType)}
              />
            ))}
          </MenuContent>
        </Menu>
        <Menu>
          <MenuTrigger label="Union type" />
          <MenuContent>
            <MenuRadioGroup value={unionType ?? QueryUnionType.AND} onValueChange={setUnionType}>
              <MenuRadioItem label="AND" value={QueryUnionType.AND} />
              <MenuRadioItem label="OR" value={QueryUnionType.OR} />
            </MenuRadioGroup>
          </MenuContent>
        </Menu>
        {relationDirection && (
          <Menu>
            <MenuTrigger label="Direction" />
            <MenuContent>
              <MenuRadioGroup value={relationDirection} onValueChange={setRelationDirection}>
                <MenuRadioItem label={QueryRelationDirection.BOTH} value={QueryRelationDirection.BOTH} />
                <MenuRadioItem label={QueryRelationDirection.RIGHT} value={QueryRelationDirection.RIGHT} />
                <MenuRadioItem label={QueryRelationDirection.LEFT} value={QueryRelationDirection.LEFT} />
              </MenuRadioGroup>
            </MenuContent>
          </Menu>
        )}
        <Menu>
          <MenuTrigger label="Entity statistics" />
          <MenuContent>
            <MenuLabel label="Total node count" />
            <MenuItem disabled label={String(activeQuery?.graphCounts?.nodeCounts[props.node?.id + '_count'] ?? 'unknown')}></MenuItem>
          </MenuContent>
        </Menu>
        <MenuSeparator />
        <MenuItem label="Remove" className="text-danger" onClick={() => removeNode()} />
      </MenuContent>
    </Menu>
  );
};
