import { Button, Input, Popover, PopoverContent, PopoverTrigger, Tooltip, TooltipContent, TooltipTrigger } from '@/lib/components';

// import { wsAddQuery, wsManualQueryRequest } from '@/lib/data-access/broker';
import { addError } from '@/lib/data-access/store/configSlice';
import { addNewQuery, clearQB, setQuerybuilderSettings } from '@/lib/data-access/store/sessionSlice';
import { useAppDispatch } from '../../data-access';
// import { ManualQueryDialog } from './ManualQueryDialog';
import { wsAddQuery } from '@/lib/data-access/broker';
import { QueryMLDialog } from './querysidepanel/QueryMLDialog';
import { QuerySettings } from './querysidepanel/QuerySettings';
// ----- Begin Button Components -----

interface TooltipButtonProps {
  tooltip: string;
  iconComponent: string;
  onClick: (e: React.MouseEvent) => void;
  variantType?: 'primary' | 'secondary' | 'danger';
  variant?: 'solid' | 'outline' | 'ghost';
  size?: 'xs' | '3xs';
  disabled?: boolean;
  className?: string;
  disableTooltip?: boolean;
}

export const TooltipButton = ({
  tooltip,
  iconComponent,
  onClick,
  variantType = 'secondary' as 'primary' | 'secondary' | 'danger',
  variant = 'ghost',
  size = 'xs',
  disabled = false,
  className = '',
  disableTooltip = false,
}: TooltipButtonProps) => (
  <Tooltip>
    <TooltipTrigger>
      <Button
        variantType={variantType}
        variant={variant}
        size={size}
        iconComponent={iconComponent}
        onClick={onClick}
        disabled={disabled}
        className={className}
      />
    </TooltipTrigger>
    <TooltipContent disabled={disableTooltip}>
      <p>{tooltip}</p>
    </TooltipContent>
  </Tooltip>
);

interface PopoverButtonProps extends TooltipButtonProps {
  popoverContent: React.ReactNode;
}

export const PopoverButton = ({
  tooltip,
  iconComponent,
  onClick,
  popoverContent,
  variantType = 'secondary',
  variant = 'ghost',
  size = 'xs',
  disabled = false,
  className = '',
  disableTooltip = false,
}: PopoverButtonProps) => (
  <Popover>
    <PopoverTrigger disabled={disabled}>
      <Tooltip>
        <TooltipTrigger>
          <Button
            variantType={variantType}
            variant={variant}
            size={size}
            iconComponent={iconComponent}
            onClick={onClick}
            disabled={disabled}
            className={className}
          />
        </TooltipTrigger>
        <TooltipContent disabled={disableTooltip}>
          <p>{tooltip}</p>
        </TooltipContent>
      </Tooltip>
    </PopoverTrigger>
    <PopoverContent>{popoverContent}</PopoverContent>
  </Popover>
);

interface AddQueryButtonProps {
  disabled: boolean;
  saveStateId: string;
}

export const AddQueryButton = ({ disabled, saveStateId }: AddQueryButtonProps) => {
  const dispatch = useAppDispatch();

  const handleAddQuery = async () => {
    wsAddQuery({ saveStateID: saveStateId }, ({ data, status }) => {
      if (status !== 'success' || data == null || !data.id || data.id < 0) {
        console.error('Failed to add query');
        addError('Failed to add query');
        return;
      }

      console.log('Query added', data);
      dispatch(addNewQuery(data));
    });
  };

  return <TooltipButton tooltip="Add query" iconComponent="icon-[ic--baseline-add]" onClick={handleAddQuery} disabled={disabled} />;
};

interface FitViewButtonProps {
  onFitView: () => void;
}

export const FitViewButton = ({ onFitView }: FitViewButtonProps) => (
  <TooltipButton tooltip="Fit to screen" iconComponent="icon-[ic--baseline-fullscreen]" onClick={onFitView} />
);

interface ClearNodesButtonProps {
  disabled: boolean;
}

export const ClearNodesButton = ({ disabled }: ClearNodesButtonProps) => {
  const dispatch = useAppDispatch();

  const handleClearNodes = () => {
    dispatch(clearQB());
  };

  return (
    <TooltipButton tooltip="Clear query panel" iconComponent="icon-[ic--baseline-delete]" onClick={handleClearNodes} disabled={disabled} />
  );
};

interface ScreenshotButtonProps {
  onScreenshot: () => void;
}

export const ScreenshotButton = ({ onScreenshot }: ScreenshotButtonProps) => (
  <TooltipButton tooltip="Capture screen" iconComponent="icon-[ic--baseline-camera-alt]" onClick={onScreenshot} />
);

interface SettingsButtonProps {
  disabled: boolean;
}

export const SettingsButton = ({ disabled }: SettingsButtonProps) => (
  <PopoverButton
    tooltip="Query builder settings"
    iconComponent="icon-[ic--baseline-settings]"
    onClick={() => {}}
    disabled={disabled}
    className="query-settings"
    popoverContent={<QuerySettings />}
  />
);

interface RunQueryButtonProps {
  onRunQuery: (useCached: boolean) => void;
}

export const RunQueryButton = ({ onRunQuery }: RunQueryButtonProps) => (
  <TooltipButton
    tooltip="Rerun query"
    iconComponent="icon-[ic--baseline-cached]"
    onClick={() => onRunQuery(true)}
    // Note: The original had onDoubleClick which we need to handle differently
  />
);

interface LogicButtonProps {
  disabled: boolean;
  onLogic: () => void;
  isActive: boolean;
}

export const LogicButton = ({ disabled, onLogic, isActive }: LogicButtonProps) => (
  <TooltipButton
    tooltip="Logic settings"
    iconComponent="icon-[ic--baseline-difference]"
    onClick={onLogic}
    disabled={disabled}
    disableTooltip={isActive}
  />
);

interface MLButtonProps {
  disabled: boolean;
  mlEnabled: boolean;
  isActive: boolean;
  mlData: any;
}

export const MLButton = ({ disabled, mlEnabled, isActive, mlData }: MLButtonProps) => {
  const ml = mlData;

  const renderMLTooltipContent = () =>
    mlEnabled ? (
      <>
        <p className="font-bold text-base">Machine learning</p>
        <p className="mb-2">Algorithms detected the following results:</p>

        {ml.linkPrediction.enabled && ml.linkPrediction.result && (
          <>
            <p className="mt-2 font-semibold">Link prediction</p>
            <p>{ml.linkPrediction.result.length} links</p>{' '}
          </>
        )}
        {ml.centrality.enabled && Object.values(ml.centrality.result).length > 0 && (
          <>
            <p className="mt-2 font-semibold">Centrality</p>
            <p>{(Object.values(ml.centrality.result) as number[]).reduce((a, b) => b + a).toFixed(2)} sum of centers</p>
          </>
        )}
        {ml.communityDetection.enabled && ml.communityDetection.result && (
          <>
            <p className="mt-2 font-semibold">Community detection</p>
            <p>{ml.communityDetection.result.length} communities</p>
          </>
        )}
        {ml.shortestPath.enabled && (
          <>
            <p className="mt-2 font-semibold">Shortest path</p>
            {ml.shortestPath.result?.length > 0 && <p># of hops: {ml.shortestPath.result.length}</p>}
            {!ml.shortestPath.srcNode ? <p>Please select source node</p> : !ml.shortestPath.trtNode && <p>Please select target node</p>}
          </>
        )}
      </>
    ) : (
      <p>Machine learning</p>
    );

  return (
    <PopoverButton
      tooltip="Machine learning"
      iconComponent="icon-[ic--baseline-lightbulb]"
      onClick={() => {}}
      variantType={mlEnabled ? 'primary' : 'secondary'}
      variant={mlEnabled ? 'outline' : 'ghost'}
      disabled={disabled}
      disableTooltip={isActive}
      popoverContent={<QueryMLDialog />}
    />
  );
};

interface LimitButtonProps {
  disabled: boolean;
  activeQuery: any;
  resultSize: number;
  totalSize: number;
}

export const LimitButton = ({ disabled, activeQuery, resultSize, totalSize }: LimitButtonProps) => {
  const dispatch = useAppDispatch();
  const limitReached = activeQuery.settings.limit <= resultSize;

  const renderLimitTooltipContent = () => (
    <>
      <p className="font-bold text-base">Limit</p>
      <p>Limits the number of edges retrieved from the database.</p>
      <p>Required to manage performance.</p>
      <p className={`font-semibold${limitReached ? ' text-primary-200' : ''}`}>
        Fetched {resultSize} out of {totalSize} nodes
      </p>
    </>
  );

  const handleLimitChange = (value: any) => {
    dispatch(setQuerybuilderSettings({ settings: { limit: Number(value) } }));
  };

  return (
    <Popover placement="bottom">
      <PopoverTrigger>
        <Tooltip>
          <TooltipTrigger>
            <Button
              variantType={limitReached ? 'primary' : 'secondary'}
              variant={limitReached ? 'outline' : 'ghost'}
              size="xs"
              disabled={disabled}
              iconComponent="icon-[ic--baseline-filter-alt]"
            />
          </TooltipTrigger>
          <TooltipContent>{renderLimitTooltipContent()}</TooltipContent>
        </Tooltip>
      </PopoverTrigger>
      <PopoverContent>
        <Input
          type="number"
          size="sm"
          label="Limit"
          value={activeQuery.settings.limit}
          lazy
          onChange={handleLimitChange}
          className={`w-24${limitReached ? ' border-danger-600' : ''}`}
          containerClassName="p-2"
        />
      </PopoverContent>
    </Popover>
  );
};

interface ManualQueryButtonProps {
  disabled: boolean;
}

// export const ManualQueryButton = ({ disabled }: ManualQueryButtonProps) => {
//   const handleManualQuery = (query: string): void => {
//     wsManualQueryRequest({ query });
//   };

//   return (
//     <PopoverButton
//       tooltip="Manual Query"
//       iconComponent="icon-[ic--baseline-search]"
//       onClick={() => {}}
//       disabled={disabled}
//       popoverContent={<ManualQueryDialog onSubmit={handleManualQuery} />}
//     />
//   );
// };

// ----- End Button Components -----

// Also export all interfaces for external use
export type {
  AddQueryButtonProps,
  ClearNodesButtonProps,
  FitViewButtonProps,
  LimitButtonProps,
  LogicButtonProps,
  //   ManualQueryButtonProps,
  MLButtonProps,
  PopoverButtonProps,
  RunQueryButtonProps,
  ScreenshotButtonProps,
  SettingsButtonProps,
  TooltipButtonProps,
};
