import { Button } from '@/lib/components';
import React, { useState } from 'react';
import { visualizationColors } from 'ts-common';

type QueryInputCypherProps = {
  cypher: string | undefined;
  onCypherChange: (value: string) => void;
  onRunQuery: () => void;
};

type ViewMode = 'raw' | 'readable';

const highlightColor = visualizationColors.GPCat.colors[14][0];

export const QueryInputCypher: React.FC<QueryInputCypherProps> = ({ cypher, onCypherChange, onRunQuery }) => {
  const [viewMode, setViewMode] = useState<ViewMode>('raw');

  const formatCypherQuery = (query: string): React.ReactNode => {
    if (!query) return null;

    // Split the query by common Cypher keywords
    const keywords = [
      'MATCH',
      'WHERE',
      'RETURN',
      'ORDER BY',
      'LIMIT',
      'WITH',
      'CREATE',
      'MERGE',
      'DELETE',
      'SET',
      'REMOVE',
      'OPTIONAL MATCH',
    ];

    // Create a regex pattern that matches any of the keywords
    const keywordPattern = new RegExp(`(${keywords.join('|')})`, 'gi');

    // Split the query by keywords and preserve the keywords
    const parts = query.split(keywordPattern).filter(x => x != '');

    return (
      <div className="font-mono text-sm">
        {parts.map((part, index) => {
          // Check if this part is a keyword
          const isKeyword = keywords.some(keyword => part.trim().toUpperCase() === keyword);

          return (
            <div
              key={index}
              style={{ color: isKeyword ? highlightColor : 'black' }}
              className={`${isKeyword ? 'font-bold' : ''} ${isKeyword ? 'ml-0' : 'ml-8'} mb-1`}
            >
              {part}
            </div>
          );
        })}
      </div>
    );
  };

  return (
    <div className="bg-neutral-50 h-full p-2 flex flex-col">
      <div className="flex justify-between items-center mb-3">
        <div>
          <h3 className="font-semibold">Cypher query</h3>
          <p className="text-neutral-500 text-sm">
            Below you find the Cypher query as executed by the backend. You can edit directly, but changes to the query are not persisted.
          </p>
        </div>
        <div className="flex space-x-2">
          <Button variantType="secondary" variant={viewMode === 'raw' ? 'solid' : 'outline'} onClick={() => setViewMode('raw')} size="sm">
            Raw Edit
          </Button>
          <Button
            variantType="secondary"
            variant={viewMode === 'readable' ? 'solid' : 'outline'}
            onClick={() => setViewMode('readable')}
            size="sm"
          >
            Readable
          </Button>
        </div>
      </div>

      <div className="flex-grow overflow-auto border rounded">
        {viewMode === 'raw' ? (
          <textarea className="w-full h-full p-2 font-mono text-sm" value={cypher} onChange={e => onCypherChange(e.target.value)} />
        ) : (
          <div className="p-2 h-full overflow-auto">{formatCypherQuery(cypher || '')}</div>
        )}
      </div>

      <div className="mt-3">
        <Button variantType="primary" variant="solid" onClick={onRunQuery}>
          Run
        </Button>
      </div>
    </div>
  );
};
