import { ControlContainer, TooltipProvider } from '@/lib/components';
import { Tabs } from '@/lib/components/tabs';
import { wsUpdateQuery } from '@/lib/data-access/broker';
import { addError } from '@/lib/data-access/store/configSlice';
import objectHash from 'object-hash';
import { useEffect, useMemo, useRef } from 'react';
import Sortable from 'sortablejs';
import { useActiveQuery, useActiveSaveState, useAppDispatch, useGraphQuery, useML } from '../../../data-access';
import { clearQB, reorderQueryState, setQueryName } from '../../../data-access/store/sessionSlice';
import {
  AddQueryButton,
  ClearNodesButton,
  FitViewButton,
  LimitButton,
  LogicButton,
  MLButton,
  RunQueryButton,
  ScreenshotButton,
  SettingsButton,
} from '../QueryBuilderControlButtons';
import QueryBuilderTab from './QueryBuilderTab';

export type QueryBuilderToggleSettings = 'settings' | 'ml' | 'logic' | 'relatedNodes' | undefined;

export type QueryBuilderNavProps = {
  toggleSettings: QueryBuilderToggleSettings;
  onFitView: () => void;
  onRunQuery: (useCached: boolean) => void;
  onScreenshot: () => void;
  onLogic: () => void;
};

export const QueryBuilderNav = (props: QueryBuilderNavProps) => {
  const dispatch = useAppDispatch();
  const activeQuery = useActiveQuery();
  const ss = useActiveSaveState();
  const graphQuery = useGraphQuery();
  const resultSize = useMemo((): number => {
    if (!graphQuery) return 0;
    return graphQuery.graph.nodes.length;
  }, [graphQuery]);
  const totalSize = useMemo((): number => {
    if (!activeQuery || !graphQuery || !ss) return 0;

    const nodeCounts = activeQuery.graph.nodes
      .filter(x => x.attributes.type == 'entity')
      .map(x => graphQuery.graphCounts?.nodeCounts[`${x.key}_count`] ?? 0);

    return nodeCounts.reduce((a, b) => a + b, 0);
  }, [graphQuery]);
  const ml = useML();

  /**
   * Clears all nodes in the graph.
   */
  function clearAllNodes(): void {
    dispatch(clearQB());
  }

  const tabsRef = useRef<HTMLDivElement | null>(null);
  useEffect(() => {
    if (!ss || !tabsRef.current) return;
    const sortable = new Sortable(tabsRef.current, {
      animation: 150,
      draggable: '[data-type="tab"]',
      ghostClass: 'bg-secondary-300',
      dragClass: 'bg-secondary-100',
      onEnd: evt => {
        if (evt.oldIndex != null && evt.newIndex != null && evt.oldIndex !== evt.newIndex) {
          dispatch(
            reorderQueryState({
              oldIndex: evt.oldIndex,
              newIndex: evt.newIndex,
            }),
          );
        }
      },
    });

    const sortedQueries = ss.queryStates.openQueryArray
      .filter(query => query.id != null)
      .sort((a, b) => {
        return a.order < b.order ? -1 : 1;
      });
    sortable.sort(sortedQueries.map(x => String(x.id)));

    return () => {
      sortable.destroy();
    };
  }, [ss ? objectHash(Object.fromEntries(ss?.queryStates.openQueryArray.map(x => [x.id, x.order]))) : null]);

  const mlEnabled = ml.linkPrediction.enabled || ml.centrality.enabled || ml.communityDetection.enabled || ml.shortestPath.enabled;

  // const handleManualQuery = (query: string): void => {
  //   wsManualQueryRequest({ query: query });
  // };

  function updateQueryName(queryId: number, name: string): void {
    if (!ss) return;
    const query = ss.queryStates.openQueryArray.find(q => q.id === queryId);
    if (!query) return;
    wsUpdateQuery(
      {
        saveStateID: ss.id,
        query: { ...query, name: name },
      },
      ({ status }) => {
        if (status !== 'success') {
          addError('Failed to update query');
        }

        dispatch(setQueryName({ queryId, name }));
      },
    );
  }

  if (!ss || !activeQuery) {
    console.debug('No active query found in query nav');
    return null;
  }

  // Get sorted queries
  const sortedQueries = ss.queryStates.openQueryArray.filter(query => query.id != null).sort((a, b) => (a.order < b.order ? -1 : 1));

  // Check if there's more than one query tab
  const hasMultipleQueries = sortedQueries.length > 1;
  const canWrite = !!ss?.authorization.database?.W;
  const canQuery = !!ss?.authorization.query?.W;

  return (
    <div className="sticky shrink-0 top-0 left-0 right-0 flex items-stretch justify-start h-7 bg-secondary-100 border-b border-secondary-200 max-w-full">
      <div className="flex items-center px-2">
        <h1 className="text-xs font-semibold text-secondary-600 truncate">Query builder</h1>
      </div>
      <div className="flex items-center px-0.5 gap-1 border-l border-secondary-200">
        <TooltipProvider>
          <AddQueryButton disabled={!canWrite} saveStateId={ss.id} />
        </TooltipProvider>
      </div>
      <Tabs ref={tabsRef} className={`-my-px overflow-x-auto overflow-y-hidden no-scrollbar divide-x divide-secondary-200 border-x`}>
        {sortedQueries.map((query, i) => (
          <QueryBuilderTab
            key={i}
            query={query}
            index={i}
            activeQueryId={activeQuery?.id || -1}
            isActive={query.id === activeQuery?.id}
            canWrite={canWrite}
            saveStateId={ss.id}
            isLastTab={!hasMultipleQueries}
            onUpdateQueryName={updateQueryName}
          />
        ))}
      </Tabs>
      <div className="sticky right-0 px-0.5 ml-auto items-center flex truncate">
        <ControlContainer>
          <TooltipProvider>
            <FitViewButton onFitView={props.onFitView} />
            <ClearNodesButton disabled={!canQuery} />
            <ScreenshotButton onScreenshot={props.onScreenshot} />
            <SettingsButton disabled={!canQuery} />
            <RunQueryButton onRunQuery={props.onRunQuery} />
            <LogicButton disabled={!canQuery} onLogic={props.onLogic} isActive={props.toggleSettings === 'logic'} />
            <MLButton disabled={!canQuery} mlEnabled={mlEnabled} isActive={props.toggleSettings === 'ml'} mlData={ml} />
            <LimitButton disabled={!canQuery} activeQuery={activeQuery} resultSize={resultSize} totalSize={totalSize} />
          </TooltipProvider>
        </ControlContainer>
      </div>
    </div>
  );
};
