import React, { useEffect, useRef, useState } from 'react';
import Plot from 'react-plotly.js';
import { prepareDataOrchestrator } from '../prepareData/prepareDataOrchestrator';
import { plotTypeOptions, Vis1DInputData } from './types';

interface Vis1DChartProps {
  title: string;
  plotType: (typeof plotTypeOptions)[number];
  xData: Vis1DInputData;
  yData: Vis1DInputData;
  zData: Vis1DInputData;
  showAxis: boolean;
  stackVariable: boolean;
  groupDataByTime?: string;
  groupAggregation?: string;
}

export const Vis1DChart: React.FC<Vis1DChartProps> = ({
  title,
  plotType,
  xData,
  yData,
  zData,
  showAxis,
  stackVariable,
  groupDataByTime,
  groupAggregation,
}) => {
  const internalRef = useRef<HTMLDivElement>(null);
  const [divSize, setDivSize] = useState({ width: 0, height: 0 });

  useEffect(() => {
    const handleResize = () => {
      if (internalRef.current) {
        const { width, height } = internalRef.current.getBoundingClientRect();
        setDivSize({ width, height });
      }
    };

    handleResize();
    window.addEventListener('resize', handleResize);
    if (internalRef.current) {
      new ResizeObserver(handleResize).observe(internalRef.current);
    }

    return () => {
      window.removeEventListener('resize', handleResize);
    };
  }, []);

  const { data, layout } = prepareDataOrchestrator(
    plotType,
    xData,
    yData,
    zData,
    showAxis,
    stackVariable,
    groupDataByTime,
    groupAggregation,
  );

  return (
    <div className="h-full w-full flex items-center justify-center overflow-hidden relative" ref={internalRef}>
      <Plot
        data={data}
        config={{
          responsive: true,
          scrollZoom: false,
          displayModeBar: false,
          displaylogo: false,
        }}
        layout={{
          ...layout,
          title: {
            text: title,
            font: {
              size: 16,
              color: 'black',
            },
            x: 0.5,
            y: 0.9,
          },
          width: divSize.width,
          height: divSize.height,
          dragmode: false,
        }}
      />
    </div>
  );
};
