import { Vis1DInputData } from '../componentsVIs1D/types';
import { getCommonLayout, getCSSVariableHSL, sharedTickFont } from '../prepareLayout/LayoutCommon';
import { PlotType } from 'plotly.js';
import { scaleOrdinal, scaleQuantize } from 'd3';
import { visualizationColors } from 'ts-common';
import { groupByTime } from './utilsTime';

export const preparePlotDataAndLayoutBar = (
  xData: Vis1DInputData,
  yData: Vis1DInputData,
  zData: Vis1DInputData,
  showAxis: boolean,
  groupDataByTime?: string,
) => {
  const primaryColor = getCSSVariableHSL('--clr-sec--400');
  const mainColors = visualizationColors.GPCat.colors[14];

  // data preparation
  let xValues: (string | number)[] = [];
  let yValues: (string | number)[] = [];

  if (groupDataByTime !== undefined) {
    if (yData.data && yData.data.length !== 0) {
      const { xValuesGrouped, yValuesGrouped } = groupByTime(xData.data as string[], groupDataByTime, yData.data);
      xValues = xValuesGrouped;
      yValues = yValuesGrouped.flat();
    } else {
      const { xValuesGrouped, yValuesGrouped } = groupByTime(xData.data as string[], groupDataByTime);
      xValues = xValuesGrouped;
      yValues = yValuesGrouped.flat();
    }
  } else {
    if (xData.data.length !== 0 && yData && yData.data.length !== 0) {
      xValues = xData.data;
      yValues = yData.data;
    } else if (xData.data.length !== 0 && (!yData || yData.data.length === 0)) {
      xValues = xData.data;
      yValues = xData.data.map((_, index) => index + 1);
    } else if (xData.data.length === 0 && yData && yData.data.length !== 0) {
      xValues = yData.data.map((_, index) => index + 1);
      yValues = yData.data;
    }
  }

  let colorScale: any;
  let colorDataZ: string[] = [];
  let colorbar: any = {};

  if (zData.data && zData.data.length > 0 && zData.type === 'number') {
    const mainColorsSeq = visualizationColors.GPSeq.colors[9];
    //const numericZAxisData = zData.data.filter((item): item is number => typeof item === 'number');
    const zMin = zData.data.reduce((min, val) => (val < min ? val : min), zData.data[0]);
    const zMax = zData.data.reduce((max, val) => (val > max ? val : max), zData.data[0]);

    // !TODO: option to have a linear or quantize scale
    colorScale = scaleQuantize<string>()
      .domain([zMin as number, zMax as number])
      .range(mainColorsSeq);

    colorDataZ = zData.data?.map(item => colorScale(item) || primaryColor);

    colorbar = {
      title: 'Color Legend',
      tickvals: [zMin, zMax],
      ticktext: [`${zMin}`, `${zMax}`],
    };
  } else {
    const uniqueZAxisData = Array.from(new Set(zData.data));

    if (zData.data && uniqueZAxisData) {
      colorScale = scaleOrdinal<string>().domain(uniqueZAxisData.map(String)).range(mainColors);

      colorDataZ = zData.data?.map(item => colorScale(String(item)) || primaryColor);
      const sortedDomain = uniqueZAxisData.sort();
      colorbar = {
        title: 'Color Legend',
        tickvals: sortedDomain,
        ticktext: sortedDomain.map(val => String(val)),
        tickmode: 'array',
      };
    }
  }
  //

  const data: Plotly.Data[] = (() => {
    if (xData.type === 'string' && groupDataByTime === undefined) {
      const frequencyMap = xData.data.reduce(
        (acc, item) => {
          acc[item] = (acc[item] || 0) + 1;
          return acc;
        },
        {} as Record<string, number>,
      );

      const sortedEntries = Object.entries(frequencyMap).sort((a, b) => b[1] - a[1]);
      const sortedLabels = sortedEntries.map(([label]) => String(label));
      const sortedFrequencies = sortedEntries.map(([, frequency]) => frequency);

      // Fix Y-axis range rounding issue
      const maxYValue = Math.ceil(Math.max(...sortedFrequencies) / 10) * 10;
      const yAxisRange = [0, maxYValue];

      return [
        {
          type: 'bar' as PlotType,
          x: sortedLabels,
          y: sortedFrequencies,
          marker: {
            color: colorDataZ?.length ? colorDataZ : primaryColor,
          },
          customdata: sortedLabels,
          hovertemplate: '<b>%{customdata}</b>: %{y}<extra></extra>',
        },
      ];
    } else if (xData.type === 'date') {
      return [
        {
          type: 'bar' as PlotType,
          x: xValues,
          y: yValues,
          marker: {
            color: colorDataZ?.length ? colorDataZ : primaryColor,
          },
          customdata: xValues,
          hovertemplate: '<b>%{customdata}</b>: %{y}<extra></extra>',
        },
      ];
    } else {
      return [
        {
          type: 'bar' as PlotType,
          x: xValues,
          y: yValues,
          marker: { color: primaryColor },
          customdata: xValues,
          hovertemplate: '<b>%{customdata}</b>: %{y}<extra></extra>',
        },
      ];
    }
  })();

  const layout: Partial<Plotly.Layout> = getCommonLayout({
    showAxis,
    xAxisLabel: xData.label,
    yAxisLabel: yData.label,
    xValues,
    yValues,
    plotType: 'line',
    xAxisData: xValues,
    sortedLabels: [],
    sharedTickFont,
  });

  return { data, layout };
};
