import { Vis1DInputData } from '../componentsVIs1D/types';
import { getCommonLayout, getCSSVariableHSL, sharedTickFont } from '../prepareLayout/LayoutCommon';
import { PlotType } from 'plotly.js';
import { scaleOrdinal } from 'd3';
import { visualizationColors } from 'ts-common';

export const preparePlotDataAndLayoutHistogramCount = (
  xData: Vis1DInputData,
  yData: Vis1DInputData,
  zData: Vis1DInputData,
  showAxis: boolean,
  stackVariable: boolean,
) => {
  const primaryColor = getCSSVariableHSL('--clr-sec--400');
  const mainColors = visualizationColors.GPCat.colors[14];

  const isCategorical = typeof xData.data[0] === 'string';

  const frequencyMap = xData.data.reduce(
    (acc, item, index) => {
      if (!acc[item]) acc[item] = { count: 0, zValueCounts: {} };
      acc[item].count += (yData.data?.[index] as number) ?? 1;
      if (zData.data) {
        const zValue = zData.data[index] || '(Empty)';
        acc[item].zValueCounts[zValue] = (acc[item].zValueCounts[zValue] || 0) + 1;
      }
      return acc;
    },
    {} as Record<string, { count: number; zValueCounts: Record<string, number> }>,
  );

  const sortedCategories = Object.entries(frequencyMap).sort((a, b) => b[1].count - a[1].count);
  const sortedLabels = sortedCategories.map(([label]) => label);

  let traces;

  if (isCategorical) {
    if (zData.data && zData.data.length > 0) {
      const uniqueZValues = Array.from(new Set(zData.data.map(String)));
      const colorScale = scaleOrdinal<string>().domain(uniqueZValues).range(mainColors);

      const groupedTraces: Record<string, { x: string[]; y: number[] }> = {};

      sortedCategories.forEach(([label, data]) => {
        Object.entries(data.zValueCounts).forEach(([zValue, count]) => {
          const color = colorScale(zValue);
          if (!groupedTraces[color]) {
            groupedTraces[color] = { x: [], y: [] };
          }
          groupedTraces[color].x.push(label);
          groupedTraces[color].y.push(stackVariable ? (count / data.count) * 100 : count);
        });
      });

      traces = Object.entries(groupedTraces).map(([color, traceData]) => ({
        type: 'bar' as PlotType,
        x: traceData.x,
        y: traceData.y,
        name: uniqueZValues.find(z => colorScale(z) === color) || '(Unknown)',
        marker: { color },
        hovertemplate: '<b>X: %{x}</b><br><b>Y: %{y}</b><extra></extra>',
        ...(stackVariable ? { stackgroup: 'one' } : {}),
      }));
    } else {
      traces = [
        {
          type: 'bar' as PlotType,
          x: sortedLabels,
          y: sortedCategories.map(([_, data]) => data.count),
          marker: { color: primaryColor },
        },
      ];
    }
  } else {
    traces = [
      {
        type: 'histogram' as PlotType,
        x: xData.data,
        marker: { color: primaryColor },
      },
    ];
  }

  const layout: Partial<Plotly.Layout> = getCommonLayout({
    showAxis,
    xAxisLabel: xData.label,
    yAxisLabel: yData.label,
    xValues: xData.data,
    yValues: yData.data,
    plotType: 'bar',
    xAxisData: xData.data,
    sortedLabels,
    sharedTickFont,
  });

  return { data: traces, layout };
};
