import { Vis1DInputData } from '../componentsVIs1D/types';
import { getCommonLayout, getCSSVariableHSL, sharedTickFont } from '../prepareLayout/LayoutCommon';
import { PlotType } from 'plotly.js';
import { scaleOrdinal, scaleQuantize } from 'd3';
import { visualizationColors } from 'ts-common';
import { computeStringTickValues } from '../prepareData/utils';

export const preparePlotDataAndLayoutScatter = (xData: Vis1DInputData, yData: Vis1DInputData, zData: Vis1DInputData, showAxis: boolean) => {
  const primaryColor = getCSSVariableHSL('--clr-sec--400');
  const mainColors = visualizationColors.GPCat.colors[14];
  const lengthLabelsX = 7; // !TODO computed number of elements based
  const lengthLabelsY = 8; // !TODO computed number of elements based

  // data preparation
  let xValues: (string | number)[] = [];
  let yValues: (string | number)[] = [];

  if (xData.data.length !== 0 && yData && yData.data.length !== 0) {
    xValues = xData.data;
    yValues = yData.data;
  } else if (xData.data.length !== 0 && (!yData || yData.data.length === 0)) {
    xValues = xData.data;
    yValues = xData.data.map((_, index) => index + 1);
  } else if (xData.data.length === 0 && yData && yData.data.length !== 0) {
    xValues = yData.data.map((_, index) => index + 1);
    yValues = yData.data;
  }

  let colorScale: any;
  let colorDataZ: string[] = [];
  let colorbar: any = {};

  if (zData.data && zData.data.length > 0 && zData.type === 'number') {
    const mainColorsSeq = visualizationColors.GPSeq.colors[9];
    //const numericZAxisData = zData.data.filter((item): item is number => typeof item === 'number');
    const zMin = zData.data.reduce((min, val) => (val < min ? val : min), zData.data[0]);
    const zMax = zData.data.reduce((max, val) => (val > max ? val : max), zData.data[0]);

    // !TODO: option to have a linear or quantize scale
    colorScale = scaleQuantize<string>()
      .domain([zMin as number, zMax as number])
      .range(mainColorsSeq);

    colorDataZ = zData.data?.map(item => colorScale(item) || primaryColor);

    colorbar = {
      title: 'Color Legend',
      tickvals: [zMin, zMax],
      ticktext: [`${zMin}`, `${zMax}`],
    };
  } else {
    const uniqueZAxisData = Array.from(new Set(zData.data));

    if (zData.data && uniqueZAxisData) {
      colorScale = scaleOrdinal<string>().domain(uniqueZAxisData.map(String)).range(mainColors);

      colorDataZ = zData.data?.map(item => colorScale(String(item)) || primaryColor);
      const sortedDomain = uniqueZAxisData.sort();
      colorbar = {
        title: 'Color Legend',
        tickvals: sortedDomain,
        ticktext: sortedDomain.map(val => String(val)),
        tickmode: 'array',
      };
    }
  }

  const data: Plotly.Data[] = [
    {
      type: 'scatter' as PlotType,
      x: xValues,
      y: yValues,
      mode: 'markers' as const,
      marker: {
        color: zData.data && zData.data.length > 0 ? colorDataZ : primaryColor,
        size: 7,
      },
      customdata:
        xValues.length === 0
          ? yValues.map(y => `Y: ${y}`)
          : yValues.length === 0
            ? xValues.map(x => `X: ${x}`)
            : xValues.map((x, index) => {
                const zValue = zData.data && zData.data.length > 0 ? zData.data[index] : null;
                return zValue ? `X: ${x} | Y: ${yValues[index]} | Color: ${zValue}` : `X: ${x} | Y: ${yValues[index]}`;
              }),
      hovertemplate: '<b>%{customdata}</b><extra></extra>',
    },
  ];

  let truncatedXLabels: string[] = [];
  let truncatedYLabels: string[] = [];

  if (typeof xValues[0] === 'string') {
    truncatedXLabels = computeStringTickValues(xValues, 2, lengthLabelsX);
  }

  if (typeof yValues[0] === 'string') {
    truncatedYLabels = computeStringTickValues(yValues, 2, lengthLabelsY);
  }

  const layout: Partial<Plotly.Layout> = getCommonLayout({
    showAxis,
    xAxisLabel: xData.label,
    yAxisLabel: yData.label,
    xValues,
    yValues,
    plotType: 'line',
    xAxisData: xValues,
    sortedLabels: [],
    sharedTickFont,
  });

  return { data, layout };
};
