import { GraphQueryResult } from '@/lib/data-access';
import { Vis1DInputData, VariableSelection } from '../componentsVIs1D/types';
import { Vis1DProps } from '../componentsVIs1D/types';

export const computeStringTickValues = (xValues: any[], maxTicks: number, maxLabelLength: number): any[] => {
  const truncatedValues = xValues.map(label => (label.length > maxLabelLength ? `${label.slice(0, maxLabelLength)}…` : label));

  return truncatedValues;
};

export const getAttributeValuesSimple = (query: GraphQueryResult, selectedEntity: string, data: VariableSelection): Vis1DInputData => {
  if (!selectedEntity || !data.name || data.name === ' ') {
    return { data: [], type: data.type, label: data.name as string };
  }

  const attValues = query.graph.nodes
    .filter(item => item.label === selectedEntity)
    .map(item => {
      return item.attributes && data.name in item.attributes && item.attributes[data.name] !== ''
        ? (item.attributes[data.name] as string | number)
        : 'NoData';
    });

  return {
    data: attValues,
    type: data.type,
    label: data.name as string,
  };
};

export const getAttributeValuesDegree = (
  query: GraphQueryResult,
  selectedEntity: string,
): { xAxisData: Vis1DInputData; yAxisData: Vis1DInputData } => {
  if (!selectedEntity) {
    return {
      xAxisData: { data: [], type: 'string', label: 'axisx' as string },
      yAxisData: { data: [], type: 'number', label: 'axisy' as string },
    };
  }

  const nodesSelected = query.graph.nodes.filter(item => item.label === selectedEntity);

  const nodeDegreeMap = nodesSelected.reduce(
    (acc, node) => {
      const nodeId = node._id;
      acc[nodeId] = query.graph.edges.reduce((degree, edge) => {
        if (edge.from === nodeId || edge.to === nodeId) {
          degree += 1;
        }
        return degree;
      }, 0);
      return acc;
    },
    {} as Record<string, number>,
  );

  const sortedNodeDegrees = Object.entries(nodeDegreeMap).sort((a, b) => b[1] - a[1]);

  const yAxisData = {
    data: sortedNodeDegrees.map(item => item[1]),
    type: 'number',
    label: 'Degree' as string,
  };

  const xAxisData = {
    data: sortedNodeDegrees.map(item => item[0]),
    type: 'string',
    label: 'Ids',
  };

  return { xAxisData, yAxisData };
};

export const getAllAxisData = (query: GraphQueryResult, settings: Vis1DProps) => {
  let xAxisData: Vis1DInputData = {
    data: [],
    label: '',
    type: '',
  };
  let yAxisData: Vis1DInputData = {
    data: [],
    label: '',
    type: '',
  };
  let zAxisData: Vis1DInputData = {
    data: [],
    label: '',
    type: '',
  };

  if (settings.groupAggregation === 'count') {
    xAxisData = getAttributeValuesSimple(query, settings.selectedEntity, settings.xAxisLabel);
    yAxisData = getAttributeValuesSimple(query, settings.selectedEntity, settings.yAxisLabel);
    zAxisData = getAttributeValuesSimple(query, settings.selectedEntity, settings.zAxisLabel);
  } else if (settings.groupAggregation === 'degree') {
    const degreeData = getAttributeValuesDegree(query, settings.selectedEntity);
    xAxisData = degreeData.xAxisData;
    yAxisData = degreeData.yAxisData;
  }

  return { xAxisData, yAxisData, zAxisData };
};
